﻿using Eneter.Messaging.EndPoints.Rpc;
using Eneter.Messaging.EndPoints.TypedMessages;
using Eneter.Messaging.MessagingSystems.MessagingSystemBase;
using Eneter.Messaging.MessagingSystems.SynchronousMessagingSystem;
using Eneter.Messaging.MessagingSystems.TcpMessagingSystem;
using Eneter.Messaging.Nodes.ChannelWrapper;

namespace ServerCalculator2
{
    // Input data for messaging calculator service
    public class CalculatorInputData
    {
        public double Number1 { get; set; }
        public double Number2 { get; set; }
    }

    // Output result from the messaging calculator service
    public class CalculatorOutputData
    {
        public double Result { get; set; }
    }

	// Interface for RPC calculator service
	public interface ICalculator
	{
		double Sum(double a, double b);
		double Subtract(double a, double b);
		double Multiplay(double a, double b);
		double Divide(double a, double b);
	}

	// Implementation of RPC service.
	internal class Calculator : ICalculator
	{
		public double Sum(double a, double b)
		{
			return a + b;
		}

		public double Subtract(double a, double b)
		{
			return a - b;
		}

		public double Multiplay(double a, double b)
		{
			return a * b;
		}

		public double Divide(double a, double b)
		{
			return a / b;
		}
	}

    internal class CalculatorService
    {
		// Unwrapps messages from the input channel and forwards them
		// to corresponding output channels.
		private IDuplexChannelUnwrapper myDuplexChannelUnwrapper;

		// Paticular messaging services listening to requests which will be forwarded from
		// the channel unwrapper.
		private IDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData> mySumReceiver;
		private IDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData> mySubtractReceiver;
		private IDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData> myMultiplyReceiver;
		private IDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData> myDivideReceiver;

		// RPC service listening to requests which will be forwarded from the channel unwrapper.
		private IRpcService<ICalculator> myRpcCalculator;

		// Calculator service logic.
		private Calculator myCalculator = new Calculator();

        public CalculatorService()
        {
            // Internal messaging used for messaging between channel unwrapper
            // and typed message receivers.
			IMessagingSystemFactory anInternalMessaging = new SynchronousMessagingSystemFactory();

            // All messages are received via one channel. So we must provide "unwrapper" forwarding incoming messages
            // to correct receivers.
            IChannelWrapperFactory aChannelWrapperFactory = new ChannelWrapperFactory();
            myDuplexChannelUnwrapper = aChannelWrapperFactory.CreateDuplexChannelUnwrapper(anInternalMessaging);

            // Factory to create message receivers.
            IDuplexTypedMessagesFactory aMessageReceiverFactory = new DuplexTypedMessagesFactory();
            
            // Create receiver to sum two numbers.
            mySumReceiver = aMessageReceiverFactory.CreateDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData>();
            mySumReceiver.MessageReceived += SumCmd; // attach method handling the request
			// attach the input channel to get messages from unwrapper
			mySumReceiver.AttachDuplexInputChannel(anInternalMessaging.CreateDuplexInputChannel("Sum"));

            // Receiver to subtract two numbers.
            mySubtractReceiver = aMessageReceiverFactory.CreateDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData>();
            mySubtractReceiver.MessageReceived += SubCmd; // attach method handling the request
            mySubtractReceiver.AttachDuplexInputChannel(anInternalMessaging.CreateDuplexInputChannel("Sub")); // attach the input channel to get messages from unwrapper

            // Receiver for multiply two numbers.
            myMultiplyReceiver = aMessageReceiverFactory.CreateDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData>();
            myMultiplyReceiver.MessageReceived += MulCmd; // attach method handling the request
            myMultiplyReceiver.AttachDuplexInputChannel(anInternalMessaging.CreateDuplexInputChannel("Mul")); // attach the input channel to get messages from unwrapper

            // Receiver for divide two numbers.
            myDivideReceiver = aMessageReceiverFactory.CreateDuplexTypedMessageReceiver<CalculatorOutputData, CalculatorInputData>();
            myDivideReceiver.MessageReceived += DivCmd; // attach method handling the request
			myDivideReceiver.AttachDuplexInputChannel(anInternalMessaging.CreateDuplexInputChannel("Div")); // attach the input channel to get messages from unwrapper

			// Connect RPC with the channel unwrapper.
			IRpcFactory anRpcFactory = new RpcFactory();
			myRpcCalculator = anRpcFactory.CreateSingleInstanceService<ICalculator>(myCalculator);
			myRpcCalculator.AttachDuplexInputChannel(anInternalMessaging.CreateDuplexInputChannel("Rpc"));
        }


        public void Start()
        {
            // We use TCP based messaging.
            IMessagingSystemFactory aServiceMessagingSystem = new TcpMessagingSystemFactory();
            IDuplexInputChannel anInputChannel = aServiceMessagingSystem.CreateDuplexInputChannel("tcp://127.0.0.1:8091/");

            // Attach the input channel to the unwrapper and start to listening.
            myDuplexChannelUnwrapper.AttachDuplexInputChannel(anInputChannel);
        }

        public void Stop()
        {
            // Detach the input channel from the unwrapper and stop listening.
            // Note: It releases listening threads.
            myDuplexChannelUnwrapper.DetachDuplexInputChannel();
        }

        // It is called when a request to sum two numbers was received.
        private void SumCmd(object sender, TypedRequestReceivedEventArgs<CalculatorInputData> e)
        {
            // Get input data.
            CalculatorInputData anInputData = e.RequestMessage;

            // Calculate output result.
            CalculatorOutputData aReturn = new CalculatorOutputData();
            aReturn.Result = myCalculator.Sum(anInputData.Number1, anInputData.Number2);

            // Response result to the client.
            mySumReceiver.SendResponseMessage(e.ResponseReceiverId, aReturn);
        }

        // It is called when a request to subtract two numbers was received.
        private void SubCmd(object sender, TypedRequestReceivedEventArgs<CalculatorInputData> e)
        {
            // Get input data.
            CalculatorInputData anInputData = e.RequestMessage;

            // Calculate output result.
            CalculatorOutputData aReturn = new CalculatorOutputData();
            aReturn.Result = myCalculator.Subtract(anInputData.Number1, anInputData.Number2);

            // Response result to the client.
            mySubtractReceiver.SendResponseMessage(e.ResponseReceiverId, aReturn);
        }
        

        // It is called when a request to multiply two numbers was received.
        private void MulCmd(object sender, TypedRequestReceivedEventArgs<CalculatorInputData> e)
        {
            // Get input data.
            CalculatorInputData anInputData = e.RequestMessage;

            // Calculate output result.
            CalculatorOutputData aReturn = new CalculatorOutputData();
            aReturn.Result = myCalculator.Multiplay(anInputData.Number1, anInputData.Number2);

            // Response result to the client.
            myMultiplyReceiver.SendResponseMessage(e.ResponseReceiverId, aReturn);
        }

        // It is called when a request to divide two numbers was received.
        private void DivCmd(object sender, TypedRequestReceivedEventArgs<CalculatorInputData> e)
        {
            // Get input data.
            CalculatorInputData anInputData = e.RequestMessage;

            // Calculate output result.
            CalculatorOutputData aReturn = new CalculatorOutputData();
            aReturn.Result = myCalculator.Divide(anInputData.Number1, anInputData.Number2);

            // Response result to the client.
            myDivideReceiver.SendResponseMessage(e.ResponseReceiverId, aReturn);
        }
    }
}
